
package walpo.negocio;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Scanner;
import walpo.modelo.CabeceraFinal;
import walpo.modelo.CabeceraLog;
import walpo.modelo.LineaFinal;
import walpo.modelo.LineaLog;

/**
 *
 * @author Antonio López
 */
public class TratarArchivo {

    private File archivoIntermedioGenerado;
    private File archivoFinalGenerado;
    private TratarLinea tratarLinea;

    /** Lista temporal que contiene las líneas del archivo de log procesado */
    private ArrayList<LineaLog> logAcceso;
    /** Lista que contiene, de forma única, las URL de los recursos accedidos registrados en el archivo de log procesado */
    private ArrayList<String> listaRecursosSolicitados;
    /** Lista con las secuencias de navegación obtenidas a partir del archivo intermedio obtenido de procesar el fichero de log */
    private ArrayList<LineaFinal> listaSecuenciasNavegacion;

    /** Número total de líneas del archivo inicial. */
    private int totalLineas;
    /** Número total de líneas que han sido descartadas durante el proceso del archivo. */
    private int totalLineasDescartadas;
    /** Formato de codificación del fichero de log leido. */
    private static final String CODIFICACION_UTF8 = "UTF-8";
    /** Sufijo para el nombre del fichero intermedio */
    private static final String SUFIJO_FICHERO_INTERMEDIO = "_INTERMEDIO";
    /** Sufijo para el nombre del fichero final */
    private static final String SUFIJO_FICHERO_FINAL = "_FINAL";
    /** Valor de la extensión para ficheros CSV */
    private static final String EXTENSION_CSV = ".csv";

    /** Constante para habilitar (true) o deshabilitar (false) el filtrado por número de accesos a los recursos */
    private static final boolean FILTRADO_POR_NUMERO_ACCESOS_RECURSO = true;
    /** Número mínimo de accesos de un recurso para no descartarlo de la lista de recursos accedidos */
    private static final int NUMERO_MINIMO_ACCESOS_RECURSO = 5;
    /** Número maximo de accesos de un recurso para no descartarlo de la lista de recursos accedidos.
        Modificar este parámetro si el fichero de log abarca un periodo de tiempo muy prolongado. */
    private static final int NUMERO_MAXIMO_ACCESOS_RECURSO = 1500;
    /** Número mínimo de accesos realizados por una IP para no descartarla */
    private static final int NUMERO_MINIMO_ACCESOS_IP = 5;

    public TratarArchivo(File archivo) throws IOException {

        //Inicializar las variables
        this.logAcceso = new ArrayList<LineaLog>();
        this.tratarLinea = new TratarLinea();
        this.totalLineas = 0;
        this.totalLineasDescartadas = 0;

        //Procesar el fichero y cargar su contenido en una estructura en memoria
        this.leerArchivo(archivo);
        //Dar nombre al fichero intermedio tomando como referencia el nombre del fichero inicial
        this.identificarArchivoGenerado(archivo, false, EXTENSION_CSV);
        //Dar nombre al fichero f tomando como referencia el nombre del fichero inicial
        this.identificarArchivoGenerado(archivo, true, EXTENSION_CSV);
    }

    /**
     * Leer el archivo pasado como parámetro y almacenar su contenido en una lista temporal en memoria.
     *
     * @param archivo
     * @throws FileNotFoundException
     * @throws IOException
     */
    private void leerArchivo(File archivo) throws IOException {

        FileInputStream fis = new FileInputStream(archivo);
        Scanner scanner = new Scanner(fis, CODIFICACION_UTF8);

        //Recorrer el archivo leyendo su contenido línea a línea
        while (scanner.hasNextLine()){

             //Procesar la línea leida y crear el objeto que la representa
            LineaLog lineaLog = this.tratarLinea.getLineaLog(scanner.nextLine());

            //Incorporar la línea leida como objeto a la lista temporal
            //Si una línea es erronea, no se interrumpe la ejecución, pero no se agrega a la lista y por tanto queda descartada
            if (lineaLog != null) { this.logAcceso.add(lineaLog); this.totalLineas++; }
        }

        //Cerrar los objetos abiertos
        scanner.close();
        fis.close();
    }

    /**
     * Eliminar aquellas líneas que no cumplan las condiciones necesarias para el estudio.
     */
    public void eliminarLineasNoNecesarias() {

        //Recorrer línea a linea la lista temporal descartando aquellas que no cumplan las condiciones necesarias
        Iterator<LineaLog> iLogAcceso = this.logAcceso.iterator();
        while (iLogAcceso.hasNext()) {

            LineaLog lineaLog = iLogAcceso.next();

            //a. Determinar si la línea refiere a un protocolo no válido
            if (!this.tratarLinea.validarProtocolo(lineaLog)) { iLogAcceso.remove(); totalLineasDescartadas++; continue; }
            //b. Determinar si la línea refiere a un recurso no válido
            if (!this.tratarLinea.validarRecursoAccedido(lineaLog)) { iLogAcceso.remove(); totalLineasDescartadas++; continue; }
            //c. Determinar si la línea ha sido generada por un sistema automatizado en vez de por un usuario
            if (!this.tratarLinea.validarSistemaAutomatizado(lineaLog)) { iLogAcceso.remove(); totalLineasDescartadas++; continue; }
            //d. Determinar si el recurso solicitado en la línea tratada contiene un identificador de sesión
            //ATENCIÓN!: dado que descartar accesos a recursos cuya URL no tenía ID de sesión hacía perder mucha información,
            //se ha decidido no omitirlos en el proceso. En su lugar, si el ID de sesión acaba tomando valor NULL, mandará la IP
            //Como sugerencia para los logs del Campus Virtual de la UOC, debería ponerse el ID de sesión en todos los accesos a recursos
//            if (!this.tratarLinea.validarIdentificadorSesion(lineaLog)) { iLogAcceso.remove(); totalLineasDescartadas++; continue; }
            //e. Determinar si la línea contiene un código de respuesta no válido
            if (!this.tratarLinea.validarCodigoRespuesta(lineaLog)) { iLogAcceso.remove(); totalLineasDescartadas++; continue; }
            //f. Determinar si la línea refiere a una dirección IP no admitida
            if (!this.tratarLinea.validarDireccionIp(lineaLog)) { iLogAcceso.remove(); totalLineasDescartadas++; continue; }
        }
    }

    /**
     * Para cada una de las líneas, determinar el identificador de sesión a partir de la URL del recurso solicitado.
     */
    public void tratarSesiones() {

        //Recorrer línea a linea la lista temporal y procesar los identificadores de sesión
        for (LineaLog lineaLog : this.logAcceso) {
            this.tratarLinea.setIdentificadorSesion(lineaLog);
        }
    }
   
    /**
     * Para cada una de las líneas, extraer los parámetros de la URL del recurso solicitado.
     */
    public void tratarUrlRecursoSolicitado() {

        //Recorrer línea a linea la lista temporal y procesar la URL del recurso solicitado
        for (LineaLog lineaLog : this.logAcceso) {
            this.tratarLinea.tratarUrlRecursoSolicitado(lineaLog);
        }
    }

    /**
     * Transformar el archivo leido a formato CSV (Comma-Separated Values) generando
     * un nuevo archivo en dicho formato e incluyendo la correspondiente cabecera.
     * @throws IOException
     */
    public void transformarCsvIntermedio() throws IOException {

        PrintWriter pw = new PrintWriter(new FileWriter(this.archivoIntermedioGenerado));

        //Escribir la cabecera (para identificar los elementos separados por comas)
        this.escribirArchivo(pw, CabeceraLog.getCabeceraLogParcial(LineaLog.CARACTER_COMA));

        //Escribir las líneas de contenido que interesa para el estudio
        for (LineaLog lineaLog: this.logAcceso) {
            this.escribirArchivo(pw, lineaLog.getLineaLogParcial(LineaLog.CARACTER_COMA));
        }

        //Cerrar el archivo
        this.finalizarEscrituraArchivo(pw);
    }

    /**
     * Dadas las líneas que se han incluido en el fichero intermedio, extraer los recursos solicitados
     * para que figuren de forma única en una lista en memoria.
     */
//    public void identificarRecursosSolicitados() {
//
//        this.listaRecursosSolicitados = new ArrayList<String>();
//        //Recorrer línea a linea la lista temporal y procesar la URL del recurso solicitado
//        for (LineaLog lineaLog : this.logAcceso) {
//
//            String recursoSolicitado = lineaLog.getRecursoSolicitado();
//
//            if (!listaRecursosSolicitados.contains(recursoSolicitado)) {
//                this.listaRecursosSolicitados.add(recursoSolicitado);
//            }
//        }
//    }

    /**
     * Dadas las líneas que se han incluido en el fichero intermedio, extraer los recursos solicitados
     * para que figuren de forma única en una lista en memoria.
     */
    public void identificarRecursosSolicitados() {

        HashMap<String, Integer> listaRecursos = new HashMap<String, Integer>();

        //Recorrer línea a linea la lista temporal y procesar la URL del recurso solicitado
        for (LineaLog lineaLog : this.logAcceso) {

            String recursoSolicitado = lineaLog.getRecursoSolicitado();

            if (listaRecursos.containsKey(recursoSolicitado)) {
                listaRecursos.put(recursoSolicitado, listaRecursos.get(recursoSolicitado) + 1);
            } else {
                listaRecursos.put(recursoSolicitado, 1);
            }
        }

        this.generarListaRecursosSolicitados(listaRecursos);
    }

    /**
     * Filtra el conjunto de recursos solicitados en función del número de accesos realizado.
     * Los recursos con un número de accesos por debajo o por encima del límite indicado son descartados.
     * Este filtrado es necesario a efectos de rendimiento y optimización de memoria; en caso de querer deshabilitarlo,
     * fijar a 'false' el valor de la constante FILTRADO_POR_NUMERO_ACCESOS_RECURSO.
     * @param listaRecursos
     */
    private void generarListaRecursosSolicitados(HashMap<String, Integer> listaRecursos) {

        this.listaRecursosSolicitados = new ArrayList<String>();

        Iterator it = listaRecursos.entrySet().iterator();
        while (it.hasNext()) {

            Map.Entry<String, Integer> e = (Map.Entry) it.next();

            if (FILTRADO_POR_NUMERO_ACCESOS_RECURSO) { //Si se encuentra habilitado el filtrado de recursos por número mínimo de accesos...

                if (e.getValue().compareTo(NUMERO_MINIMO_ACCESOS_RECURSO) > 0 &&
                    e.getValue().compareTo(NUMERO_MAXIMO_ACCESOS_RECURSO) < 0) { this.listaRecursosSolicitados.add(e.getKey()); }

            } else { this.listaRecursosSolicitados.add(e.getKey()); }
        }
    }


    /**
     * A partir del mismo conjunto de líneas que se han incluido en el fichero intermedio, determinar
     * las secuencias de navegación.
     * Una secuencia de navegación es el conjunto de recursos accedidos por una dirección IP para
     * un mismo ID de sesión. De esta forma, una misma dirección IP puede contar con varias secuencias
     * de navegación si ha accedido con distintos ID de sesión.
     */
    public void determinarSecuenciasNavegacion() {

        this.listaSecuenciasNavegacion = new ArrayList<LineaFinal>();

        //Recorrer línea a línea la lista temporal y procesar la URL del recurso solicitado
        for (LineaLog lineaLog : this.logAcceso) {
            
            boolean existe = false;

            //Dada la línea de log tratada, determinar si ya ha sido considerada en la lista de secuencias de navegación
            for (LineaFinal lineaFinal : this.listaSecuenciasNavegacion) {

                //Se considera una misma línea cuando al comparar la línea del fichero de log con la línea final
                //cuando hay coincidencia de direcciones IP y de IDs de sesión.
                if (lineaLog.getDireccionIP().equals(lineaFinal.getDireccionIP())) {

                    if (lineaLog.getIdSesion().equals(lineaFinal.getIdSesion())) {

                        //Si la línea se ha considerado en la lista de secuencias de navegación,
                        //se retoma la lista de recursos que figura en la misma
                        //y se completa con el recurso que figura en la línea de log tratada
                        lineaFinal.inicializarListaRecursos(this.listaRecursosSolicitados, lineaLog.getRecursoSolicitado());

                        existe = true;
                    }                    
                }                
            }

            //Si se ha determinado que la línea no había sido considerada aún en la lista de secuencias de navegación,
            //se crea y se añade a la lista
            if (!existe) {

                LineaFinal lineaFinal = new LineaFinal(lineaLog.getDireccionIP(), lineaLog.getIdSesion(),
                        this.listaRecursosSolicitados, lineaLog.getRecursoSolicitado());

                this.listaSecuenciasNavegacion.add(lineaFinal);
            }
        }

        if (FILTRADO_POR_NUMERO_ACCESOS_RECURSO) { //Si se encuentra habilitado el filtrado de recursos por número mínimo de accesos...
            //Descartar direcciones IP cuya secuencia de navegación tome valor 'F' en todos los casos
            this.descartarDireccionesIpSinAccesos();
            this.descartarRecursosNoAccedidos();
        }
    }

    /**
     * Si se encuentra habilitado el filtrado de recursos accedidos en función del número de accesos,
     * existirán direcciones IP cuya secuencia de navegación toma valores 'F' (sin acceso) para todos
     * los recursos considerados. Por lo tanto, es necesario eliminar dichas direcciones IP y así no falsear
     * los resultados obtenidos.
     */
    public void descartarDireccionesIpSinAccesos() {       

        Iterator<LineaFinal> iSecuenciasNavegacion = this.listaSecuenciasNavegacion.iterator();
        while (iSecuenciasNavegacion.hasNext()) {

            LineaFinal lineaFinal = iSecuenciasNavegacion.next();

            boolean eliminarLinea = true;

            int contadorAccesos = 0; //Almacena el número de accesos (recurso = 'T') ha realizado una dirección IP

            Iterator it = lineaFinal.getListaRecursos().entrySet().iterator();
            while (it.hasNext()) {

                Map.Entry<String, String> e = (Map.Entry) it.next();
                if (e.getValue().equals(LineaFinal.ACCEDIDO)) { eliminarLinea = false; contadorAccesos++; }
            }

            if (eliminarLinea || contadorAccesos < NUMERO_MINIMO_ACCESOS_IP) { iSecuenciasNavegacion.remove(); continue; }
        }
    }

    /**
     * Dado que distorsionan los resultados del estudio, descartar para cada secuencia de navegación, todos aquellos recursos
     * que tras los diferentes filtrados realizados figuren como no accedidos por todas las direcciones IP (valor 'F').
     */
    public void descartarRecursosNoAccedidos() {
        
        ArrayList<String> listaRecursosNoDescartados = new ArrayList<String>();

        //Recorrer la lista de secuencias de navegación para determinar que recursos han tenido accesos (valor 'T')
        Iterator<LineaFinal> iSecuenciasNavegacion = this.listaSecuenciasNavegacion.iterator();
        while (iSecuenciasNavegacion.hasNext()) {

            LineaFinal lineaFinal = iSecuenciasNavegacion.next();

            Iterator it = lineaFinal.getListaRecursos().entrySet().iterator();
            while (it.hasNext()) {

                Map.Entry<String, String> e = (Map.Entry) it.next();
                if (e.getValue().equals(LineaFinal.ACCEDIDO) && !listaRecursosNoDescartados.contains(e.getKey())) {
                    listaRecursosNoDescartados.add(e.getKey());
                }
            }
        }

        //Recorrer la lista de secuencias de navegación para eliminar aquellos recursos sin accesos (valor 'F').
        //Los recursos descartados serán aquellos que no figuren en la lista de recursos no descartados.
        //La operación se repite para cada una de las secuencias de navegación donde, para cada una de ellas,
        //se eliminan los mismos recursos.
        iSecuenciasNavegacion = this.listaSecuenciasNavegacion.iterator(); //Reiniciar el iterador
        while (iSecuenciasNavegacion.hasNext()) {

            LineaFinal lineaFinal = iSecuenciasNavegacion.next();

            Iterator it = lineaFinal.getListaRecursos().entrySet().iterator();
            while (it.hasNext()) {

                Map.Entry<String, String> e = (Map.Entry) it.next();
                if (!listaRecursosNoDescartados.contains(e.getKey())) { it.remove(); continue; }
            }
        }

        //Recorrer la lista de recursos accedidos y eliminar aquellos descartados durante la ejecución de este método;
        //es decir, aquellos que no figuran en la lista de recursos no descartados
        Iterator<String> iRecursosSolicitados = this.listaRecursosSolicitados.iterator();
        while (iRecursosSolicitados.hasNext()) {

            String recursoSolicitado = iRecursosSolicitados.next();

            if (!listaRecursosNoDescartados.contains(recursoSolicitado)) { iRecursosSolicitados.remove(); continue; }
        }
    }


    /**
     * Generar el archivo final formato CSV (Comma-Separated Values) incluyendo su
     * correspondiente cabecera.
     * @throws IOException
     */
    public void transformarCsvFinal() throws IOException {

        PrintWriter pw = new PrintWriter(new FileWriter(this.archivoFinalGenerado));

        //Escribir la cabecera (para identificar los elementos separados por comas)
        this.escribirArchivo(pw, CabeceraFinal.getCabeceraFinal(LineaFinal.CARACTER_COMA, this.listaRecursosSolicitados));

        //Escribir las líneas de contenido que interesa para el estudio
        for (LineaFinal lineaFinal: this.listaSecuenciasNavegacion) {
            this.escribirArchivo(pw, lineaFinal.getLineaFinal(LineaFinal.CARACTER_COMA));
        }

        //Cerrar el archivo
        this.finalizarEscrituraArchivo(pw);
    }

    /**
     * Escribir el el archivo destino la línea indicada como parámetro.
     * @param pw
     * @param linea
     * @return
     */
    private PrintWriter escribirArchivo(PrintWriter pw, String linea) {

        pw.println(linea);
        pw.flush();

        return pw;
    }

    /**
     * Cerrar el archivo que se ha estado procesando.
     * @param pw
     */
    private void finalizarEscrituraArchivo(PrintWriter pw) {
        pw.close();
    }

    /**
     * Definir el nombre y ubicación del fichero que se va a generar.
     * @param archivo
     * @param archivoFinal (false para el fichero intermedio, true para el archivo final)
     * @param extension
     */
    private void identificarArchivoGenerado(File archivo, boolean archivoFinal, String extension) {

        //Esta variable almacena ruta, nombre y extensión del archivo
        String nombreArchivo = archivo.getAbsolutePath();

        //Determinar la posición del punto para encontrar la extensión del archivo
        int posicionPunto = nombreArchivo.lastIndexOf('.');

        if (0 < posicionPunto && posicionPunto <= nombreArchivo.length() - 2 ) {

            //Almacenar ruta y nombre (sin la extensión)
            String nombreNuevoArchivo = nombreArchivo.substring(0, posicionPunto);

            //Definir el nuevo archivo dándole un nuevo nombre y conservando ruta y extensión
            if (!archivoFinal) { this.archivoIntermedioGenerado = new File(nombreNuevoArchivo + SUFIJO_FICHERO_INTERMEDIO + extension); }
            else { this.archivoFinalGenerado = new File(nombreNuevoArchivo + SUFIJO_FICHERO_FINAL + extension); }
        }
    }

    /**
     * Obtener el número total de líneas del archivo inicial.
     */
    public int getTotalLineas() {
        return this.totalLineas;
    }

    /**
     * Obtener el número total de líneas que han sido descartadas tras procesar el archivo.
     */
    public int getTotalLineasDescartadas() {
        return this.totalLineasDescartadas;
    }

    /**
     * Obtener el número total de líneas que han sido aceptadas tras procesar el archivo.
     */
    public int getTotalLineasAceptadas() {
        return this.getTotalLineas() - this.getTotalLineasDescartadas();
    }
}
