
package walpo.negocio;

import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import walpo.auxiliares.ConfigurarLog;
import walpo.modelo.LineaLog;

/**
 *
 * @author Antonio López
 */
public class TratarLinea {
    
    /** Número de elementos esperado en cada línea del log inicial */
    private static final int NUMERO_ELEMENTOS_LINEA = 10;
    /** Formato que debe cumplir cada una de las líneas de log.
     * Se espera que una línea de log tenga este formato:
     * [<fecha y hora registro>] <dirección IP> "<método petición> <ruta recurso solicitado> <protocolo>" <código respuesta>
     * "<ruta origen>" "<navegador>" <n.º bytes obtenido> <p1>
     * Cada grupo de la expresión regular, corresponde a un elemento de la lína de log.
     */
    private static final String FORMATO_LINEA = "\\[([\\w:/]+\\s[+\\-]\\d{4})\\]* ([\\d.]+) \"(.+) (/.+) (.+)\" (\\d{3}) \"(.+)\" \"(.+)\" (\\d+) (\\d+)";

    /** Formato que indica que el protocolo que aparece en la línea de log es válido */
    //Para el estudio, se considera únicamente el protocolo HTTP
    private static final String PROTOCOLO_VALIDO = "HTTP.+";
    /** Conjunto de recursos a los que no debe referir una línea de log */
    //Extensiones de archivos que no interesan para el estudio
    private static final String[] RECURSOS_NO_VALIDOS = {"js", "JS", "jpg", "JPG", "gif", "GIF", "png", "PNG", "bmp", "BMP", "ico", "ICO", "css", "CSS",
                    "text", "swf", "SWF"};
    /** Conjunto de recursos a los que puede referir una línea de log.
     * Son recursos que pueden dar conflicto con las expresiones regulares que tratan la constante RECURSOS_NO_VALIDOS
     */
    private static final String RECURSOS_VALIDOS = ".+(\\.jsp|\\.JSP).*";
    /** Formato que indica que el navegador que aparece en la línea no es válido; es decir, el valor que toma el elemento
     correspondiente de la línea de log cuando no se ha podido determinar el navegador desde el que se ha realizado el acceso.
     */
    private static final String NAVEGADOR_NO_VALIDO = "-";
    /** Formato que indica que el recurso solicitado en una línea contiene un identificador de sesión */
    //Los identificadores de sesión del Campus Virtual de la UOC tienen un tamaño de 128 caracteres
    private static final String IDENTIFICADOR_SESION_VALIDO = ".+[\\?|&]s=(.{128}).*";
    /** Formato que indica que el que la línea contienen un código de respuesta válido */
    private static final String CODIGO_RESPUESTA_VALIDO = "2\\d{2}+";
    /** Conjunto de direcciones IP no admitidas */
    private static final String[] DIRECCIONES_IP_NO_ADMITIDAS = {"192.168.1.130", "192.168.1.137"};

    /** Patrón para extraer el valor de un identificador de sesión */
    private static final String FORMATO_IDENTIFICADOR_SESION = IDENTIFICADOR_SESION_VALIDO;
    /** Patrón para extraer la URL de un recurso solicitado sin los parámetros que la acompañan.
     Los parámetros en una URL comienzan a partir del caracter '?'; en caso de haber varios, van separados
     por el caracter '&'. En algún caso, la parte a excluir va detrás del carácter ';'.
     */
    private static final String FORMATO_RECURSO_SOLICITADO = "(.+)(;.+)|(.+?)(\\?.+)|(.+^?)(.*)";

    private String patronRecursosNoValidos;


    public TratarLinea() {

        //Definir el nivel de los mensajes de log, poner a OFF para deshabilitar completamente y agilizar la ejecución
        new ConfigurarLog(Logger.getLogger(TratarLinea.class.getName())).setLevel(Level.OFF);

        this.patronRecursosNoValidos = this.getPatronRecursosNoValidos();
    }

    /**
     * Procesar la línea leida y crear el objeto que la representa.
     * Si la línea no tiene el formato esperado, lanza una excepción.
     * @param linea
     * @return
     */
    public LineaLog getLineaLog(String linea){

        /*System.out.println("Using RE Pattern:");
        System.out.println(FORMATO_LINEA);

        System.out.println("Input line is:");
        System.out.println(linea);*/

        Pattern p = Pattern.compile(FORMATO_LINEA);
        Matcher matcher = p.matcher(linea);

        //En caso de que la línea no tenga el formato esperado, se descarta
        if (!matcher.matches() || NUMERO_ELEMENTOS_LINEA != matcher.groupCount()) {
            Logger.getLogger(TratarLinea.class.getName()).log(Level.WARNING, "La siguiente línea tiene un formato no válido: \n" + linea +
                    "\nElementos: " + matcher.groupCount());
            return null;
        }

        /*System.out.println("Date&Time: " + matcher.group(1));
        System.out.println("IP Address: " + matcher.group(2));
        System.out.println("Petición: " + matcher.group(3));
        System.out.println("Request: " + matcher.group(4));
        System.out.println("Protocol: " + matcher.group(5));
        System.out.println("Response: " + matcher.group(6));
        System.out.println("Referer: " + matcher.group(7));
        System.out.println("Browser: " + matcher.group(8));
        System.out.println("Bytes Sent: " + matcher.group(9));
        System.out.println("??: " + matcher.group(10));*/

        //Crear el objeto que almacenará la línea
        LineaLog lineaLogAcceso = new LineaLog();
        lineaLogAcceso.setFechaHora(matcher.group(1));
        lineaLogAcceso.setDireccionIP(matcher.group(2).trim());
        lineaLogAcceso.setMetodoPeticion(matcher.group(3));
        lineaLogAcceso.setRecursoSolicitado(matcher.group(4).trim());
        lineaLogAcceso.setProtocolo(matcher.group(5));
        lineaLogAcceso.setCodigoRespuesta(matcher.group(6));
        lineaLogAcceso.setRutaOrigen(matcher.group(7));
        lineaLogAcceso.setNavegador(matcher.group(8));
        lineaLogAcceso.setBytesObtenidos(matcher.group(9));
        lineaLogAcceso.setP1(matcher.group(10));

        return lineaLogAcceso;
    }

    /**
     * Determinar si la línea tratada refiere a un protocolo no válido.
     * @param lineaLog
     * @return
     */
    public boolean validarProtocolo(LineaLog lineaLog) {

        Pattern p = Pattern.compile(PROTOCOLO_VALIDO);
        Matcher matcher = p.matcher(lineaLog.getProtocolo());

        //En caso de que la línea no tenga el formato esperado, se descarta
        if (!matcher.matches()) {
            Logger.getLogger(TratarLinea.class.getName()).log(Level.WARNING, "Protocolo no válido en línea: \n" + lineaLog.getLineaLog(LineaLog.CARACTER_ESPACIO));
            return false;
        }

        return true;
    }

    /**
     * Determinar si la línea tratada refiere a un recurso válido para el estudio.
     * @param lineaLog
     * @return
     */
    public boolean validarRecursoAccedido(LineaLog lineaLog) {

        Pattern p = Pattern.compile(RECURSOS_VALIDOS);
        Matcher matcher = p.matcher(lineaLog.getRecursoSolicitado());

        //Dado que puede haber conflictos con algunos de los recursos considerados no válidos, primero hay que tratarlos.
        //El caso más claro son los JSP, sin esta línea, la expresión regular que trata los recursos no válidos
        //los descarta a la par que los ficheros JS (por empezar igual).
        if (matcher.matches()) { return true; }

        p = Pattern.compile(this.patronRecursosNoValidos);
        matcher = p.matcher(lineaLog.getRecursoSolicitado());

        //En caso de que la línea refiera a un recurso no válido, se descarta
        if (matcher.matches()) {
            Logger.getLogger(TratarLinea.class.getName()).log(Level.WARNING, "La siguiente línea refiere a un recurso no válido: \n" + lineaLog.getLineaLog(LineaLog.CARACTER_ESPACIO));
            return false;
        }

        return true;
    }

    /**
     * Determinar si la línea tratada ha sido generada por un usuario del Campus Virtual o por un sistema automatizado.
     * @param lineaLog
     * @return
     */
    public boolean validarSistemaAutomatizado(LineaLog lineaLog) {

        Pattern p = Pattern.compile(NAVEGADOR_NO_VALIDO);
        Matcher matcher = p.matcher(lineaLog.getNavegador());

        //En caso de que la línea haya sido generada por el acceso de un sistema automatizado y no de un usuario, se descarta
        if (matcher.matches()) {
            Logger.getLogger(TratarLinea.class.getName()).log(Level.WARNING, "La siguiente línea se ha generado por un acceso automatizado: \n" + lineaLog.getLineaLog(LineaLog.CARACTER_ESPACIO));
            return false;
        }

        return true;
    }

    /**
     * Determinar si el recurso solicitado en la línea tratada contiene un identificador de sesión.
     * @param lineaLog
     * @return
     */
    public boolean validarIdentificadorSesion(LineaLog lineaLog) {

        Pattern p = Pattern.compile(IDENTIFICADOR_SESION_VALIDO);
        Matcher matcher = p.matcher(lineaLog.getRecursoSolicitado());

        //En caso de que la línea no contenga un identificador de sesión en el recurso accedido, se descarta
        if (!matcher.matches()) {
            Logger.getLogger(TratarLinea.class.getName()).log(Level.WARNING, "La siguiente línea no contiene un identificador de sesión en el recurso solicitado: \n" + lineaLog.getLineaLog(LineaLog.CARACTER_ESPACIO));
            return false;
        }

        return true;
    }

    /**
     * Determinar si la línea contiene un código de respuesta válido.
     * @param lineaLog
     * @return
     */
    public boolean validarCodigoRespuesta(LineaLog lineaLog) {

        Pattern p = Pattern.compile(CODIGO_RESPUESTA_VALIDO);
        Matcher matcher = p.matcher(lineaLog.getCodigoRespuesta());

        //En caso de que la línea no contenga un identificador de sesión en el recurso accedido, se descarta
        if (!matcher.matches()) {
            Logger.getLogger(TratarLinea.class.getName()).log(Level.WARNING, "La siguiente línea contiene un código de respuesta no válido: \n" + lineaLog.getLineaLog(LineaLog.CARACTER_ESPACIO));
            return false;
        }

        return true;
    }

    /**
     * Determinar si la línea refiere a una dirección IP no admitida.
     * @param lineaLog
     * @return
     */
    public boolean validarDireccionIp(LineaLog lineaLog) {

        for (String direccionIp : DIRECCIONES_IP_NO_ADMITIDAS) {

            //En caso de que la línea refiera a una dirección IP no admitida, se descarta
            if (lineaLog.getDireccionIP().equals(direccionIp)) {
                Logger.getLogger(TratarLinea.class.getName()).log(Level.WARNING, "La siguiente línea refiere a una dirección IP no admitida: \n" + lineaLog.getLineaLog(LineaLog.CARACTER_ESPACIO));
                return false;
            }
        }

        return true;
    }

    /**
     * Extraer el identificador de sesión de la URL del recurso solicitado actualizando el objeto que representa a la línea tratada.
     * @param lineaLog
     */
    public void setIdentificadorSesion(LineaLog lineaLog){

        Pattern p = Pattern.compile(FORMATO_IDENTIFICADOR_SESION);
        Matcher matcher = p.matcher(lineaLog.getRecursoSolicitado());

        //En caso de que el identificador de sesión no tenga el formato esperado, el ID de sesión queda como NULL
        //En tratamientos de líneas con ID de sesión con valor NULL, mandará el valor de la dirección IP
        if (!matcher.matches()) {
            Logger.getLogger(TratarLinea.class.getName()).log(Level.WARNING, "La siguiente línea no tiene un identificador de sesión válido: \n" + lineaLog.getLineaLog(LineaLog.CARACTER_ESPACIO));
            lineaLog.setIdSesion(LineaLog.NO_ID_SESION);
        } else {
            lineaLog.setIdSesion(matcher.group(1));
        }
    }

    /**
     * Eliminar los parámetros de la URL del recurso solicitado actualizando el objeto que representa a la línea tratada.
     * @param lineaLog
     */
    public void tratarUrlRecursoSolicitado(LineaLog lineaLog){

        Pattern p = Pattern.compile(FORMATO_RECURSO_SOLICITADO);
        Matcher matcher = p.matcher(lineaLog.getRecursoSolicitado());

        //En caso de que la URL del recurso solicitado no tenga el formato esperado, se descarta
        if (!matcher.matches()) {
            Logger.getLogger(TratarLinea.class.getName()).log(Level.WARNING, "La siguiente línea se ha descartado por un problema en la URL del recurso: \n" + lineaLog.getLineaLog(LineaLog.CARACTER_ESPACIO));
            return;
        }

        //Inicializar el atributo en el objeto línea
        //Dado que la expresión regular tiene dos OR (representados por '|'), si la primera condición no se cumple (group(1)==NULL),
        //es necesario saltar a la siguiente (group(3)) y, como último recurso, a group(0)
        if (matcher.group(1)!=null) { lineaLog.setRecursoSolicitado(matcher.group(1)); }
        else if (matcher.group(3)!=null) { lineaLog.setRecursoSolicitado(matcher.group(3)); }
        else { lineaLog.setRecursoSolicitado(matcher.group(0)); }
    }

    /**
     * Dar forma a la expresión regular que determinará si el recurso al que refiere una línea
     * es válido para el estudio o no.
     * @return
     */
    private String getPatronRecursosNoValidos() {

        String patron = ".+(";

        for (int i = 0; i < RECURSOS_NO_VALIDOS.length; i++) {
            patron = patron + "\\." + RECURSOS_NO_VALIDOS[i];
            if (i < RECURSOS_NO_VALIDOS.length - 1) { patron+="|"; }
        }
        
        patron+=").*";

        return patron;
    }
}
